@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,14 +15,14 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """Command dispatcher for TabbedBrowser."""
 
 import os.path
 import shlex
 import functools
-import typing
+from typing import cast, Callable, Dict, Union
 
 from PyQt5.QtWidgets import QApplication, QTabBar
 from PyQt5.QtCore import pyqtSlot, Qt, QUrl, QEvent, QUrlQuery
@@ -38,7 +38,7 @@ from qutebrowser.utils import (message, usertypes, log, qtutils, urlutils,
 from qutebrowser.utils.usertypes import KeyMode
 from qutebrowser.misc import editor, guiprocess, objects
 from qutebrowser.completion.models import urlmodel, miscmodels
-from qutebrowser.mainwindow import mainwindow
+from qutebrowser.mainwindow import mainwindow, windowundo
 
 
 class CommandDispatcher:
@@ -65,7 +65,7 @@ class CommandDispatcher:
 
     def _new_tabbed_browser(self, private):
         """Get a tabbed-browser from a new window."""
-        args = QApplication.instance().arguments()
+        args = objects.qapp.arguments()
         if private and '--single-process' in args:
             raise cmdutils.CommandError("Private windows are unavailable with "
                                         "the single-process process model.")
@@ -170,7 +170,7 @@ class CommandDispatcher:
             elif mode == "stack-next":
                 tab = tab_deque.next(cur_tab)
             else:
-                raise NotImplementedError(
+                raise utils.Unreachable(
                     "Missing implementation for stack mode!")
         except IndexError:
             if not show_error:
@@ -278,7 +278,7 @@ class CommandDispatcher:
             return
 
         to_pin = not tab.data.pinned
-        self._tabbed_browser.widget.set_tab_pinned(tab, to_pin)
+        tab.set_pinned(to_pin)
 
     @cmdutils.register(instance='command-dispatcher', name='open',
                        maxsplit=0, scope='window')
@@ -329,7 +329,9 @@ class CommandDispatcher:
                         # Explicit count with a tab that doesn't exist.
                         return
                 elif curtab.navigation_blocked():
-                    message.info("Tab is pinned!")
+                    message.info("Tab is pinned! Opening in new tab.")
+                    self._tabbed_browser.tabopen(cur_url)
+
                 else:
                     curtab.load_url(cur_url)
 
@@ -421,12 +423,13 @@ class CommandDispatcher:
         newtab.data.keep_icon = True
         newtab.history.private_api.deserialize(history)
         newtab.zoom.set_factor(curtab.zoom.factor())
-        new_tabbed_browser.widget.set_tab_pinned(newtab, curtab.data.pinned)
+
+        newtab.set_pinned(curtab.data.pinned)
         return newtab
 
     @cmdutils.register(instance='command-dispatcher', scope='window',
                        maxsplit=0)
-    @cmdutils.argument('index', completion=miscmodels.other_buffer)
+    @cmdutils.argument('index', completion=miscmodels.other_tabs)
     def tab_take(self, index, keep=False):
         """Take a tab from another window.
 
@@ -439,7 +442,7 @@ class CommandDispatcher:
             raise cmdutils.CommandError("Can't take tabs when using "
                                         "windows as tabs")
 
-        tabbed_browser, tab = self._resolve_buffer_index(index)
+        tabbed_browser, tab = self._resolve_tab_index(index)
 
         if tabbed_browser is self._tabbed_browser:
             raise cmdutils.CommandError("Can't take a tab from the same "
@@ -498,7 +501,7 @@ class CommandDispatcher:
             self._tabbed_browser.close_tab(self._current_widget(),
                                            add_undo=False)
 
-    def _back_forward(self, tab, bg, window, count, forward):
+    def _back_forward(self, tab, bg, window, count, forward, index=None):
         """Helper function for :back/:forward."""
         history = self._current_widget().history
         # Catch common cases before e.g. cloning tab
@@ -512,6 +515,12 @@ class CommandDispatcher:
         else:
             widget = self._current_widget()
 
+        if count is None:
+            if index is None:
+                count = 1
+            else:
+                count = abs(history.current_idx() - index)
+
         try:
             if forward:
                 widget.history.forward(count)
@@ -522,7 +531,10 @@ class CommandDispatcher:
 
     @cmdutils.register(instance='command-dispatcher', scope='window')
     @cmdutils.argument('count', value=cmdutils.Value.count)
-    def back(self, tab=False, bg=False, window=False, count=1):
+    @cmdutils.argument('index', completion=miscmodels.back)
+    def back(self, tab: bool = False, bg: bool = False,
+             window: bool = False, count: int = None,
+             index: int = None) -> None:
         """Go back in the history of the current tab.
 
         Args:
@@ -530,12 +542,16 @@ class CommandDispatcher:
             bg: Go back in a background tab.
             window: Go back in a new window.
             count: How many pages to go back.
+            index: Which page to go back to, count takes precedence.
         """
-        self._back_forward(tab, bg, window, count, forward=False)
+        self._back_forward(tab, bg, window, count, forward=False, index=index)
 
     @cmdutils.register(instance='command-dispatcher', scope='window')
     @cmdutils.argument('count', value=cmdutils.Value.count)
-    def forward(self, tab=False, bg=False, window=False, count=1):
+    @cmdutils.argument('index', completion=miscmodels.forward)
+    def forward(self, tab: bool = False, bg: bool = False,
+                window: bool = False, count: int = None,
+                index: int = None) -> None:
         """Go forward in the history of the current tab.
 
         Args:
@@ -543,12 +559,13 @@ class CommandDispatcher:
             bg: Go forward in a background tab.
             window: Go forward in a new window.
             count: How many pages to go forward.
+            index: Which page to go forward to, count takes precedence.
         """
-        self._back_forward(tab, bg, window, count, forward=True)
+        self._back_forward(tab, bg, window, count, forward=True, index=index)
 
     @cmdutils.register(instance='command-dispatcher', scope='window')
     @cmdutils.argument('where', choices=['prev', 'next', 'up', 'increment',
-                                         'decrement'])
+                                         'decrement', 'strip'])
     @cmdutils.argument('count', value=cmdutils.Value.count)
     def navigate(self, where: str, tab: bool = False, bg: bool = False,
                  window: bool = False, count: int = 1) -> None:
@@ -573,6 +590,7 @@ class CommandDispatcher:
                   Uses the
                   link:settings{outsuffix}#url.incdec_segments[url.incdec_segments]
                   config option.
+                - `strip`: Strip query and fragment from the current URL.
 
             tab: Open in a new tab.
             bg: Open in a background tab.
@@ -584,24 +602,21 @@ class CommandDispatcher:
         widget = self._current_widget()
         url = self._current_url()
 
-        handlers = {
+        handlers: Dict[str, Callable] = {
             'prev': functools.partial(navigate.prevnext, prev=True),
             'next': functools.partial(navigate.prevnext, prev=False),
             'up': navigate.path_up,
-            'decrement': functools.partial(navigate.incdec,
-                                           inc_or_dec='decrement'),
-            'increment': functools.partial(navigate.incdec,
-                                           inc_or_dec='increment'),
-        }  # type: typing.Dict[str, typing.Callable]
+            'strip': navigate.strip,
+            'decrement': functools.partial(navigate.incdec, inc_or_dec='decrement'),
+            'increment': functools.partial(navigate.incdec, inc_or_dec='increment'),
+        }
 
         try:
             if where in ['prev', 'next']:
                 handler = handlers[where]
                 handler(browsertab=widget, win_id=self._win_id, baseurl=url,
                         tab=tab, background=bg, window=window)
-            elif where in ['up', 'increment', 'decrement']:
-                if where == 'up':
-                    url = url.adjusted(QUrl.RemoveFragment | QUrl.RemoveQuery)
+            elif where in ['up', 'increment', 'decrement', 'strip']:
                 new_url = handlers[where](url, count)
                 self._open(new_url, tab, bg, window, related=True)
             else:  # pragma: no cover
@@ -660,7 +675,7 @@ class CommandDispatcher:
 
         url = QUrl(self._current_url())
         url_query = QUrlQuery()
-        url_query_str = urlutils.query_string(url)
+        url_query_str = url.query()
         if '&' not in url_query_str and ';' in url_query_str:
             url_query.setQueryDelimiters('=', ';')
         url_query.setQuery(url_query_str)
@@ -751,7 +766,8 @@ class CommandDispatcher:
         """
         cmdutils.check_exclusive((prev, next_), 'pn')
         cur_idx = self._tabbed_browser.widget.currentIndex()
-        assert cur_idx != -1
+        if cur_idx == -1:
+            raise cmdutils.CommandError("Failed to get current tab for :tab-only")
 
         def _to_close(i):
             """Helper method to check if a tab should be closed or not."""
@@ -780,12 +796,39 @@ class CommandDispatcher:
                 text="Are you sure you want to close pinned tabs?")
 
     @cmdutils.register(instance='command-dispatcher', scope='window')
-    def undo(self):
-        """Re-open the last closed tab or tabs."""
+    @cmdutils.argument('count', value=cmdutils.Value.count)
+    @cmdutils.argument('depth', completion=miscmodels.undo)
+    def undo(self, window: bool = False,
+             count: int = None, depth: int = None) -> None:
+        """Re-open the last closed tab(s) or window.
+
+        Args:
+            window: Re-open the last closed window (and its tabs).
+            count: How deep in the undo stack to find the tab or tabs to
+                   re-open.
+            depth: Same as `count` but as argument for completion, `count`
+                   takes precedence.
+        """
+        has_depth = count is not None or depth is not None
+        if count is not None:
+            depth = count
+        elif depth is None:
+            depth = 1
+
+        if window and has_depth:
+            raise cmdutils.CommandError(
+                ":undo --window does not support a count/depth")
+
         try:
-            self._tabbed_browser.undo()
+            if window:
+                windowundo.instance.undo_last_window_close()
+            else:
+                self._tabbed_browser.undo(depth)
         except IndexError:
-            raise cmdutils.CommandError("Nothing to undo!")
+            msg = "Nothing to undo"
+            if not window and not has_depth:
+                msg += " (use :undo --window to reopen a closed window)"
+            raise cmdutils.CommandError(msg)
 
     @cmdutils.register(instance='command-dispatcher', scope='window')
     @cmdutils.argument('count', value=cmdutils.Value.count)
@@ -827,8 +870,8 @@ class CommandDispatcher:
         else:
             log.webview.debug("Last tab")
 
-    def _resolve_buffer_index(self, index):
-        """Resolve a buffer index to the tabbedbrowser and tab.
+    def _resolve_tab_index(self, index):
+        """Resolve a tab index to the tabbedbrowser and tab.
 
         Args:
             index: The [win_id/]index of the tab to be selected. Or a substring
@@ -840,7 +883,7 @@ class CommandDispatcher:
             for part in index_parts:
                 int(part)
         except ValueError:
-            model = miscmodels.buffer()
+            model = miscmodels.tabs()
             model.set_pattern(index)
             if model.count() > 0:
                 index = model.data(model.first_item())
@@ -875,9 +918,9 @@ class CommandDispatcher:
 
     @cmdutils.register(instance='command-dispatcher', scope='window',
                        maxsplit=0)
-    @cmdutils.argument('index', completion=miscmodels.buffer)
+    @cmdutils.argument('index', completion=miscmodels.tabs)
     @cmdutils.argument('count', value=cmdutils.Value.count)
-    def buffer(self, index=None, count=None):
+    def tab_select(self, index=None, count=None):
         """Select tab by index or url/title best match.
 
         Focuses window if necessary when index is given. If both index and
@@ -897,18 +940,17 @@ class CommandDispatcher:
         if count is not None:
             index = str(count)
 
-        tabbed_browser, tab = self._resolve_buffer_index(index)
+        tabbed_browser, tab = self._resolve_tab_index(index)
 
         window = tabbed_browser.widget.window()
-        window.activateWindow()
-        window.raise_()
+        mainwindow.raise_window(window)
         tabbed_browser.widget.setCurrentWidget(tab)
 
     @cmdutils.register(instance='command-dispatcher', scope='window')
     @cmdutils.argument('index', choices=['last', 'stack-next', 'stack-prev'],
                        completion=miscmodels.tab_focus)
     @cmdutils.argument('count', value=cmdutils.Value.count)
-    def tab_focus(self, index: typing.Union[str, int] = None,
+    def tab_focus(self, index: Union[str, int] = None,
                   count: int = None, no_last: bool = False) -> None:
         """Select the tab given as argument/[count].
 
@@ -952,7 +994,7 @@ class CommandDispatcher:
     @cmdutils.register(instance='command-dispatcher', scope='window')
     @cmdutils.argument('index', choices=['+', '-'])
     @cmdutils.argument('count', value=cmdutils.Value.count)
-    def tab_move(self, index: typing.Union[str, int] = None,
+    def tab_move(self, index: Union[str, int] = None,
                  count: int = None) -> None:
         """Move the current tab according to the argument and [count].
 
@@ -1017,7 +1059,8 @@ class CommandDispatcher:
             verbose: Show notifications when the command started/exited.
             output: Show the output in a new tab.
             output_messages: Show the output as messages.
-            detach: Whether the command should be detached from qutebrowser.
+            detach: Detach the command from qutebrowser so that it continues
+                    running when qutebrowser quits.
             cmdline: The commandline to execute.
             count: Given to userscripts as $QUTE_COUNT.
         """
@@ -1240,6 +1283,21 @@ class CommandDispatcher:
             raise cmdutils.CommandError("Bookmark '{}' not found!".format(url))
         message.info("Removed bookmark {}".format(url))
 
+    @cmdutils.register(instance='command-dispatcher', scope='window')
+    def bookmark_list(self, jump=False, tab=True, bg=False, window=False):
+        """Show all bookmarks/quickmarks.
+
+        Args:
+            tab: Open in a new tab.
+            bg: Open in a background tab.
+            window: Open in a new window.
+            jump: Jump to the "bookmarks" header.
+        """
+        url = QUrl('qute://bookmarks/')
+        if jump:
+            url.setFragment('bookmarks')
+        self._open(url, tab, bg, window)
+
     @cmdutils.register(instance='command-dispatcher', scope='window')
     def download(self, url=None, *, mhtml_=False, dest=None):
         """Download a given URL, or current page if no URL given.
@@ -1264,7 +1322,7 @@ class CommandDispatcher:
             if mhtml_:
                 raise cmdutils.CommandError("Can only download the current "
                                             "page as mhtml.")
-            url = urlutils.qurl_from_user_input(url)
+            url = QUrl.fromUserInput(url)
             urlutils.raise_cmdexc_if_invalid(url)
             download_manager.get(url, target=target)
         elif mhtml_:
@@ -1302,6 +1360,7 @@ class CommandDispatcher:
                       the case for QtWebKit. For QtWebEngine it may display
                       slightly different source.
                       Some JavaScript processing may be applied.
+                      Needs the optional Pygments dependency for highlighting.
         """
         tab = self._current_widget()
         try:
@@ -1390,7 +1449,7 @@ class CommandDispatcher:
         query = QUrlQuery()
         query.addQueryItem('level', level)
         if plain:
-            query.addQueryItem('plain', typing.cast(str, None))
+            query.addQueryItem('plain', cast(str, None))
 
         if logfilter:
             try:
@@ -1404,8 +1463,8 @@ class CommandDispatcher:
 
         self._open(url, tab, bg, window)
 
-    def _open_editor_cb(self, elem):
-        """Open editor after the focus elem was found in open_editor."""
+    def _edit_text_cb(self, elem):
+        """Open editor after the focus elem was found in edit_text."""
         if elem is None:
             message.error("No element focused!")
             return
@@ -1429,14 +1488,14 @@ class CommandDispatcher:
         ed.edit(text, caret_position)
 
     @cmdutils.register(instance='command-dispatcher', scope='window')
-    def open_editor(self):
+    def edit_text(self):
         """Open an external editor with the currently selected form field.
 
         The editor which should be launched can be configured via the
         `editor.command` config option.
         """
         tab = self._current_widget()
-        tab.elements.find_focused(self._open_editor_cb)
+        tab.elements.find_focused(self._edit_text_cb)
 
     def on_file_updated(self, ed, elem, text):
         """Write the editor text into the form field and clean up tempfile.
@@ -1586,10 +1645,32 @@ class CommandDispatcher:
             tab.search.prev_result()
         tab.search.prev_result(result_cb=cb)
 
+    def _jseval_cb(self, out):
+        """Show the data returned from JS."""
+        if out is None:
+            # Getting the actual error (if any) seems to be difficult.
+            # The error does end up in
+            # BrowserPage.javaScriptConsoleMessage(), but
+            # distinguishing between :jseval errors and errors from the
+            # webpage is not trivial...
+            message.info('No output or error')
+        else:
+            # The output can be a string, number, dict, array, etc. But
+            # *don't* output too much data, as this will make
+            # qutebrowser hang
+            out = str(out)
+            if len(out) > 5000:
+                out = out[:5000] + ' [...trimmed...]'
+            message.info(out)
+
     @cmdutils.register(instance='command-dispatcher', scope='window',
                        maxsplit=0, no_cmd_split=True)
-    def jseval(self, js_code: str, file: bool = False, quiet: bool = False, *,
-               world: typing.Union[usertypes.JsWorld, int] = None) -> None:
+    def jseval(self, js_code: str,
+               file: bool = False,
+               url: bool = False,
+               quiet: bool = False,
+               *,
+               world: Union[usertypes.JsWorld, int] = None) -> None:
         """Evaluate a JavaScript string.
 
         Args:
@@ -1598,33 +1679,24 @@ class CommandDispatcher:
                   If the path is relative, the file is searched in a js/ subdir
                   in qutebrowser's data dir, e.g.
                   `~/.local/share/qutebrowser/js`.
+            url: Interpret js-code as a `javascript:...` URL.
             quiet: Don't show resulting JS object.
             world: Ignored on QtWebKit. On QtWebEngine, a world ID or name to
-                   run the snippet in.
+                   run the snippet in. Predefined world names are:
+
+                      - `main` (same world as the web page's JavaScript and
+                        Greasemonkey, unless overridden via `@qute-js-world`)
+                      - `application` (used for internal qutebrowser JS code,
+                        should not be used via `:jseval` unless you know what
+                        you're doing)
+                      - `user` (currently unused)
+                      - `jseval` (used for this command by default)
         """
+        cmdutils.check_exclusive((file, url), 'fu')
+
         if world is None:
             world = usertypes.JsWorld.jseval
-
-        if quiet:
-            jseval_cb = None
-        else:
-            def jseval_cb(out):
-                """Show the data returned from JS."""
-                if out is None:
-                    # Getting the actual error (if any) seems to be difficult.
-                    # The error does end up in
-                    # BrowserPage.javaScriptConsoleMessage(), but
-                    # distinguishing between :jseval errors and errors from the
-                    # webpage is not trivial...
-                    message.info('No output or error')
-                else:
-                    # The output can be a string, number, dict, array, etc. But
-                    # *don't* output too much data, as this will make
-                    # qutebrowser hang
-                    out = str(out)
-                    if len(out) > 5000:
-                        out = out[:5000] + ' [...trimmed...]'
-                    message.info(out)
+        jseval_cb = None if quiet else self._jseval_cb
 
         if file:
             path = os.path.expanduser(js_code)
@@ -1636,6 +1708,11 @@ class CommandDispatcher:
                     js_code = f.read()
             except OSError as e:
                 raise cmdutils.CommandError(str(e))
+        elif url:
+            try:
+                js_code = urlutils.parse_javascript_url(QUrl(js_code))
+            except urlutils.Error as e:
+                raise cmdutils.CommandError(str(e))
 
         widget = self._current_widget()
         try:
