@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,31 +15,46 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """Our own QNetworkAccessManager."""
 
-import os
 import collections
-import netrc
 import html
+import dataclasses
+from typing import TYPE_CHECKING, Dict, MutableMapping, Optional, Sequence
 
-from PyQt5.QtCore import (pyqtSlot, pyqtSignal, PYQT_VERSION, QCoreApplication,
-                          QUrl, QByteArray)
-from PyQt5.QtNetwork import QNetworkAccessManager, QNetworkReply, QSslSocket
+from PyQt5.QtCore import pyqtSlot, pyqtSignal, QUrl, QByteArray
+from PyQt5.QtNetwork import (QNetworkAccessManager, QNetworkReply, QSslSocket,
+                             QSslError, QNetworkProxy)
 
 from qutebrowser.config import config
-from qutebrowser.utils import (message, log, usertypes, utils, objreg, qtutils,
-                               urlutils)
+from qutebrowser.utils import (message, log, usertypes, utils, objreg,
+                               urlutils, debug)
 from qutebrowser.browser import shared
-from qutebrowser.browser.webkit import certificateerror
+from qutebrowser.browser.network import proxy as proxymod
+from qutebrowser.extensions import interceptors
+from qutebrowser.browser.webkit import certificateerror, cookies, cache
 from qutebrowser.browser.webkit.network import (webkitqutescheme, networkreply,
                                                 filescheme)
+from qutebrowser.misc import objects
+
+if TYPE_CHECKING:
+    from qutebrowser.mainwindow import prompt
 
 
 HOSTBLOCK_ERROR_STRING = '%HOSTBLOCK%'
-ProxyId = collections.namedtuple('ProxyId', 'type, hostname, port')
-_proxy_auth_cache = {}
+_proxy_auth_cache: Dict['ProxyId', 'prompt.AuthInfo'] = {}
+
+
+@dataclasses.dataclass(frozen=True)
+class ProxyId:
+
+    """Information identifying a proxy server."""
+
+    type: QNetworkProxy.ProxyType
+    hostname: str
+    port: int
 
 
 def _is_secure_cipher(cipher):
@@ -49,23 +64,23 @@ def _is_secure_cipher(cipher):
         # https://codereview.qt-project.org/#/c/75943/
         return False
     # OpenSSL should already protect against this in a better way
-    elif cipher.keyExchangeMethod() == 'DH' and os.name == 'nt':
+    elif cipher.keyExchangeMethod() == 'DH' and utils.is_windows:
         # https://weakdh.org/
         return False
     elif cipher.encryptionMethod().upper().startswith('RC4'):
-        # http://en.wikipedia.org/wiki/RC4#Security
+        # https://en.wikipedia.org/wiki/RC4#Security
         # https://codereview.qt-project.org/#/c/148906/
         return False
     elif cipher.encryptionMethod().upper().startswith('DES'):
-        # http://en.wikipedia.org/wiki/Data_Encryption_Standard#Security_and_cryptanalysis
+        # https://en.wikipedia.org/wiki/Data_Encryption_Standard#Security_and_cryptanalysis
         return False
     elif 'MD5' in tokens:
-        # http://www.win.tue.nl/hashclash/rogue-ca/
+        # https://www.win.tue.nl/hashclash/rogue-ca/
         return False
     # OpenSSL should already protect against this in a better way
     # elif (('CBC3' in tokens or 'CBC' in tokens) and (cipher.protocol() not in
     #         [QSsl.TlsV1_0, QSsl.TlsV1_1, QSsl.TlsV1_2])):
-    #     # http://en.wikipedia.org/wiki/POODLE
+    #     # https://en.wikipedia.org/wiki/POODLE
     #     return False
     ### These things should never happen as those are already filtered out by
     ### either the SSL libraries or Qt - but let's be sure.
@@ -88,14 +103,9 @@ def _is_secure_cipher(cipher):
 
 def init():
     """Disable insecure SSL ciphers on old Qt versions."""
-    if qtutils.version_check('5.3.0'):
-        default_ciphers = QSslSocket.defaultCiphers()
-        log.init.debug("Default Qt ciphers: {}".format(
-            ', '.join(c.name() for c in default_ciphers)))
-    else:
-        # https://codereview.qt-project.org/#/c/75943/
-        default_ciphers = QSslSocket.supportedCiphers()
-        log.init.debug("Supported Qt ciphers: {}".format(
+    default_ciphers = QSslSocket.defaultCiphers()
+    log.init.vdebug(  # type: ignore[attr-defined]
+        "Default Qt ciphers: {}".format(
             ', '.join(c.name() for c in default_ciphers)))
 
     good_ciphers = []
@@ -106,9 +116,13 @@ def init():
         else:
             bad_ciphers.append(cipher)
 
-    log.init.debug("Disabling bad ciphers: {}".format(
-        ', '.join(c.name() for c in bad_ciphers)))
-    QSslSocket.setDefaultCiphers(good_ciphers)
+    if bad_ciphers:
+        log.init.debug("Disabling bad ciphers: {}".format(
+            ', '.join(c.name() for c in bad_ciphers)))
+        QSslSocket.setDefaultCiphers(good_ciphers)
+
+
+_SavedErrorsType = MutableMapping[urlutils.HostTupleType, Sequence[QSslError]]
 
 
 class NetworkManager(QNetworkAccessManager):
@@ -124,10 +138,13 @@ class NetworkManager(QNetworkAccessManager):
         _scheme_handlers: A dictionary (scheme -> handler) of supported custom
                           schemes.
         _win_id: The window ID this NetworkManager is associated with.
+                 (or None for generic network managers)
         _tab_id: The tab ID this NetworkManager is associated with.
+                 (or None for generic network managers)
         _rejected_ssl_errors: A {QUrl: [SslError]} dict of rejected errors.
         _accepted_ssl_errors: A {QUrl: [SslError]} dict of accepted errors.
         _private: Whether we're in private browsing mode.
+        netrc_used: Whether netrc authentication was performed.
 
     Signals:
         shutting_down: Emitted when the QNAM is shutting down.
@@ -139,7 +156,9 @@ class NetworkManager(QNetworkAccessManager):
         log.init.debug("Initializing NetworkManager")
         with log.disable_qt_msghandler():
             # WORKAROUND for a hang when a message is printed - See:
-            # http://www.riverbankcomputing.com/pipermail/pyqt/2014-November/035045.html
+            # https://www.riverbankcomputing.com/pipermail/pyqt/2014-November/035045.html
+            #
+            # Still needed on Qt/PyQt 5.15.2 according to #6010.
             super().__init__(parent)
         log.init.debug("NetworkManager init done")
         self.adopted_downloads = 0
@@ -147,30 +166,30 @@ class NetworkManager(QNetworkAccessManager):
         self._tab_id = tab_id
         self._private = private
         self._scheme_handlers = {
-            'qute': webkitqutescheme.QuteSchemeHandler(win_id),
-            'file': filescheme.FileSchemeHandler(win_id),
+            'qute': webkitqutescheme.handler,
+            'file': filescheme.handler,
         }
         self._set_cookiejar()
         self._set_cache()
         self.sslErrors.connect(self.on_ssl_errors)
-        self._rejected_ssl_errors = collections.defaultdict(list)
-        self._accepted_ssl_errors = collections.defaultdict(list)
+        self._rejected_ssl_errors: _SavedErrorsType = collections.defaultdict(list)
+        self._accepted_ssl_errors: _SavedErrorsType = collections.defaultdict(list)
         self.authenticationRequired.connect(self.on_authentication_required)
-        self.proxyAuthenticationRequired.connect(
-            self.on_proxy_authentication_required)
+        self.proxyAuthenticationRequired.connect(self.on_proxy_authentication_required)
+        self.netrc_used = False
 
     def _set_cookiejar(self):
         """Set the cookie jar of the NetworkManager correctly."""
         if self._private:
-            cookie_jar = objreg.get('ram-cookie-jar')
+            cookie_jar = cookies.ram_cookie_jar
         else:
-            cookie_jar = objreg.get('cookie-jar')
+            cookie_jar = cookies.cookie_jar
+        assert cookie_jar is not None
 
         # We have a shared cookie jar - we restore its parent so we don't
         # take ownership of it.
         self.setCookieJar(cookie_jar)
-        app = QCoreApplication.instance()
-        cookie_jar.setParent(app)
+        cookie_jar.setParent(objects.qapp)
 
     def _set_cache(self):
         """Set the cache of the NetworkManager correctly."""
@@ -178,10 +197,8 @@ class NetworkManager(QNetworkAccessManager):
             return
         # We have a shared cache - we restore its parent so we don't take
         # ownership of it.
-        app = QCoreApplication.instance()
-        cache = objreg.get('cache')
-        self.setCache(cache)
-        cache.setParent(app)
+        self.setCache(cache.diskcache)
+        cache.diskcache.setParent(objects.qapp)
 
     def _get_abort_signals(self, owner=None):
         """Get a list of signals which should abort a question."""
@@ -191,6 +208,7 @@ class NetworkManager(QNetworkAccessManager):
         # This might be a generic network manager, e.g. one belonging to a
         # DownloadManager. In this case, just skip the webview thing.
         if self._tab_id is not None:
+            assert self._win_id is not None
             tab = objreg.get('tab', scope='tab', window=self._win_id,
                              tab=self._tab_id)
             abort_on.append(tab.load_started)
@@ -203,7 +221,7 @@ class NetworkManager(QNetworkAccessManager):
 
     # No @pyqtSlot here, see
     # https://github.com/qutebrowser/qutebrowser/issues/2213
-    def on_ssl_errors(self, reply, errors):  # pragma: no mccabe
+    def on_ssl_errors(self, reply, errors):  # noqa: C901 pragma: no mccabe
         """Decide if SSL errors should be ignored or not.
 
         This slot is called on SSL/TLS errors by the self.sslErrors signal.
@@ -213,21 +231,23 @@ class NetworkManager(QNetworkAccessManager):
             errors: A list of errors.
         """
         errors = [certificateerror.CertificateErrorWrapper(e) for e in errors]
-        log.webview.debug("Certificate errors: {!r}".format(
+        log.network.debug("Certificate errors: {!r}".format(
             ' / '.join(str(err) for err in errors)))
         try:
-            host_tpl = urlutils.host_tuple(reply.url())
+            host_tpl: Optional[urlutils.HostTupleType] = urlutils.host_tuple(
+                reply.url())
         except ValueError:
             host_tpl = None
             is_accepted = False
             is_rejected = False
         else:
+            assert host_tpl is not None
             is_accepted = set(errors).issubset(
                 self._accepted_ssl_errors[host_tpl])
             is_rejected = set(errors).issubset(
                 self._rejected_ssl_errors[host_tpl])
 
-        log.webview.debug("Already accepted: {} / "
+        log.network.debug("Already accepted: {} / "
                           "rejected {}".format(is_accepted, is_rejected))
 
         if is_rejected:
@@ -267,30 +287,19 @@ class NetworkManager(QNetworkAccessManager):
     @pyqtSlot('QNetworkReply*', 'QAuthenticator*')
     def on_authentication_required(self, reply, authenticator):
         """Called when a website needs authentication."""
-        user, password = None, None
-        if not hasattr(reply, "netrc_used") and 'HOME' in os.environ:
-            # We'll get an OSError by netrc if 'HOME' isn't available in
-            # os.environ. We don't want to log that, so we prevent it
-            # altogether.
-            reply.netrc_used = True
-            try:
-                net = netrc.netrc(config.get('network', 'netrc-file'))
-                authenticators = net.authenticators(reply.url().host())
-                if authenticators is not None:
-                    (user, _account, password) = authenticators
-            except FileNotFoundError:
-                log.misc.debug("No .netrc file found")
-            except OSError:
-                log.misc.exception("Unable to read the netrc file")
-            except netrc.NetrcParseError:
-                log.misc.exception("Error when parsing the netrc file")
+        url = reply.url()
+        log.network.debug("Authentication requested for {}, netrc_used {}"
+                          .format(url.toDisplayString(), self.netrc_used))
 
-        if user is not None:
-            authenticator.setUser(user)
-            authenticator.setPassword(password)
-        else:
+        netrc_success = False
+        if not self.netrc_used:
+            self.netrc_used = True
+            netrc_success = shared.netrc_authentication(url, authenticator)
+
+        if not netrc_success:
+            log.network.debug("Asking for credentials")
             abort_on = self._get_abort_signals(reply)
-            shared.authentication_required(reply.url(), authenticator,
+            shared.authentication_required(url, authenticator,
                                            abort_on=abort_on)
 
     @pyqtSlot('QNetworkProxy', 'QAuthenticator*')
@@ -298,9 +307,9 @@ class NetworkManager(QNetworkAccessManager):
         """Called when a proxy needs authentication."""
         proxy_id = ProxyId(proxy.type(), proxy.hostName(), proxy.port())
         if proxy_id in _proxy_auth_cache:
-            user, password = _proxy_auth_cache[proxy_id]
-            authenticator.setUser(user)
-            authenticator.setPassword(password)
+            authinfo = _proxy_auth_cache[proxy_id]
+            authenticator.setUser(authinfo.user)
+            authenticator.setPassword(authinfo.password)
         else:
             msg = '<b>{}</b> says:<br/>{}'.format(
                 html.escape(proxy.hostName()),
@@ -338,7 +347,7 @@ class NetworkManager(QNetworkAccessManager):
 
     def set_referer(self, req, current_url):
         """Set the referer header."""
-        referer_header_conf = config.get('network', 'referer-header')
+        referer_header_conf = config.val.content.headers.referer
 
         try:
             if referer_header_conf == 'never':
@@ -355,13 +364,6 @@ class NetworkManager(QNetworkAccessManager):
             # https://www.playstation.com/ for example.
             pass
 
-    # WORKAROUND for:
-    # http://www.riverbankcomputing.com/pipermail/pyqt/2014-September/034806.html
-    #
-    # By returning False, we provoke a TypeError because of a wrong return
-    # type, which does *not* trigger a segfault but invoke our return handler
-    # immediately.
-    @utils.prevent_exceptions(False)
     def createRequest(self, op, req, outgoing_data):
         """Return a new QNetworkReply object.
 
@@ -373,60 +375,62 @@ class NetworkManager(QNetworkAccessManager):
         Return:
             A QNetworkReply.
         """
-        proxy_factory = objreg.get('proxy-factory', None)
-        if proxy_factory is not None:
-            proxy_error = proxy_factory.get_error()
+        if proxymod.application_factory is not None:
+            proxy_error = proxymod.application_factory.get_error()
             if proxy_error is not None:
                 return networkreply.ErrorNetworkReply(
                     req, proxy_error, QNetworkReply.UnknownProxyError,
                     self)
 
-        scheme = req.url().scheme()
-        if scheme in self._scheme_handlers:
-            result = self._scheme_handlers[scheme].createRequest(
-                op, req, outgoing_data)
-            if result is not None:
-                return result
-
-        for header, value in shared.custom_headers():
-            req.setRawHeader(header, value)
-
-        host_blocker = objreg.get('host-blocker')
-        if host_blocker.is_blocked(req.url()):
-            log.webview.info("Request to {} blocked by host blocker.".format(
-                req.url().host()))
+        if not req.url().isValid():
+            log.network.debug("Ignoring invalid requested URL: {}".format(
+                req.url().errorString()))
             return networkreply.ErrorNetworkReply(
-                req, HOSTBLOCK_ERROR_STRING, QNetworkReply.ContentAccessDenied,
+                req, "Invalid request URL", QNetworkReply.HostNotFoundError,
                 self)
 
+        for header, value in shared.custom_headers(url=req.url()):
+            req.setRawHeader(header, value)
+
         # There are some scenarios where we can't figure out current_url:
         # - There's a generic NetworkManager, e.g. for downloads
         # - The download was in a tab which is now closed.
         current_url = QUrl()
 
         if self._tab_id is not None:
+            assert self._win_id is not None
             try:
                 tab = objreg.get('tab', scope='tab', window=self._win_id,
                                  tab=self._tab_id)
                 current_url = tab.url()
-            except (KeyError, RuntimeError, TypeError):
+            except (KeyError, RuntimeError):
                 # https://github.com/qutebrowser/qutebrowser/issues/889
-                # Catching RuntimeError and TypeError because we could be in
-                # the middle of the webpage shutdown here.
+                # Catching RuntimeError because we could be in the middle of
+                # the webpage shutdown here.
                 current_url = QUrl()
 
-        self.set_referer(req, current_url)
+        request = interceptors.Request(first_party_url=current_url,
+                                       request_url=req.url())
+        interceptors.run(request)
+        if request.is_blocked:
+            return networkreply.ErrorNetworkReply(
+                req, HOSTBLOCK_ERROR_STRING, QNetworkReply.ContentAccessDenied,
+                self)
 
-        if PYQT_VERSION < 0x050301:
-            # WORKAROUND (remove this when we bump the requirements to 5.3.1)
-            #
-            # If we don't disable our message handler, we get a freeze if a
-            # warning is printed due to a PyQt bug, e.g. when clicking a
-            # currency on http://ch.mouser.com/localsites/
-            #
-            # See http://www.riverbankcomputing.com/pipermail/pyqt/2014-June/034420.html
-            with log.disable_qt_msghandler():
-                reply = super().createRequest(op, req, outgoing_data)
-        else:
-            reply = super().createRequest(op, req, outgoing_data)
-        return reply
+        if 'log-requests' in objects.debug_flags:
+            operation = debug.qenum_key(QNetworkAccessManager, op)
+            operation = operation.replace('Operation', '').upper()
+            log.network.debug("{} {}, first-party {}".format(
+                operation,
+                req.url().toDisplayString(),
+                current_url.toDisplayString()))
+
+        scheme = req.url().scheme()
+        if scheme in self._scheme_handlers:
+            result = self._scheme_handlers[scheme](req, op, current_url)
+            if result is not None:
+                result.setParent(self)
+                return result
+
+        self.set_referer(req, current_url)
+        return super().createRequest(op, req, outgoing_data)
