@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2018-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2018-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,18 +15,18 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """Loader for qutebrowser extensions."""
 
-import importlib.abc
 import pkgutil
 import types
-import typing
 import sys
 import pathlib
-
-import attr
+import importlib
+import argparse
+import dataclasses
+from typing import Callable, Iterator, List, Optional, Set, Tuple
 
 from PyQt5.QtCore import pyqtSlot
 
@@ -35,25 +35,22 @@ from qutebrowser.config import config
 from qutebrowser.utils import log, standarddir
 from qutebrowser.misc import objects
 
-if typing.TYPE_CHECKING:
-    import argparse
-
 
 # ModuleInfo objects for all loaded plugins
 _module_infos = []
 
 
-@attr.s
+@dataclasses.dataclass
 class InitContext:
 
     """Context an extension gets in its init hook."""
 
-    data_dir = attr.ib()  # type: pathlib.Path
-    config_dir = attr.ib()  # type: pathlib.Path
-    args = attr.ib()  # type: argparse.Namespace
+    data_dir: pathlib.Path
+    config_dir: pathlib.Path
+    args: argparse.Namespace
 
 
-@attr.s
+@dataclasses.dataclass
 class ModuleInfo:
 
     """Information attached to an extension module.
@@ -61,21 +58,18 @@ class ModuleInfo:
     This gets used by qutebrowser.api.hook.
     """
 
-    _ConfigChangedHooksType = typing.List[typing.Tuple[typing.Optional[str],
-                                                       typing.Callable]]
-
-    skip_hooks = attr.ib(False)  # type: bool
-    init_hook = attr.ib(None)  # type: typing.Optional[typing.Callable]
-    config_changed_hooks = attr.ib(
-        attr.Factory(list))  # type: _ConfigChangedHooksType
+    skip_hooks: bool = False
+    init_hook: Optional[Callable] = None
+    config_changed_hooks: List[Tuple[Optional[str], Callable]] = dataclasses.field(
+        default_factory=list)
 
 
-@attr.s
+@dataclasses.dataclass
 class ExtensionInfo:
 
     """Information about a qutebrowser extension."""
 
-    name = attr.ib()  # type: str
+    name: str
 
 
 def add_module_info(module: types.ModuleType) -> ModuleInfo:
@@ -92,7 +86,7 @@ def load_components(*, skip_hooks: bool = False) -> None:
         _load_component(info, skip_hooks=skip_hooks)
 
 
-def walk_components() -> typing.Iterator[ExtensionInfo]:
+def walk_components() -> Iterator[ExtensionInfo]:
     """Yield ExtensionInfo objects for all modules."""
     if hasattr(sys, 'frozen'):
         yield from _walk_pyinstaller()
@@ -104,7 +98,7 @@ def _on_walk_error(name: str) -> None:
     raise ImportError("Failed to import {}".format(name))
 
 
-def _walk_normal() -> typing.Iterator[ExtensionInfo]:
+def _walk_normal() -> Iterator[ExtensionInfo]:
     """Walk extensions when not using PyInstaller."""
     for _finder, name, ispkg in pkgutil.walk_packages(
             # Only packages have a __path__ attribute,
@@ -114,10 +108,15 @@ def _walk_normal() -> typing.Iterator[ExtensionInfo]:
             onerror=_on_walk_error):
         if ispkg:
             continue
+        if name == 'qutebrowser.components.adblock':
+            # WORKAROUND for packaging issues where the old adblock.py file is still
+            # lingering around.
+            log.extensions.debug("Ignoring stale 'adblock' component")
+            continue
         yield ExtensionInfo(name=name)
 
 
-def _walk_pyinstaller() -> typing.Iterator[ExtensionInfo]:
+def _walk_pyinstaller() -> Iterator[ExtensionInfo]:
     """Walk extensions when using PyInstaller.
 
     See https://github.com/pyinstaller/pyinstaller/issues/1905
@@ -125,10 +124,10 @@ def _walk_pyinstaller() -> typing.Iterator[ExtensionInfo]:
     Inspired by:
     https://github.com/webcomics/dosage/blob/master/dosagelib/loader.py
     """
-    toc = set()  # type: typing.Set[str]
+    toc: Set[str] = set()
     for importer in pkgutil.iter_importers('qutebrowser'):
         if hasattr(importer, 'toc'):
-            toc |= importer.toc
+            toc |= importer.toc  # type: ignore[union-attr]
     for name in toc:
         if name.startswith(components.__name__ + '.'):
             yield ExtensionInfo(name=name)
