@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2016-2020 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2016-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,13 +15,13 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """Showing messages above the statusbar."""
 
-import typing
+from typing import MutableSequence
 
-from PyQt5.QtCore import pyqtSlot, pyqtSignal, QTimer, Qt, QSize
+from PyQt5.QtCore import pyqtSlot, pyqtSignal, QTimer, Qt
 from PyQt5.QtWidgets import QWidget, QVBoxLayout, QLabel, QSizePolicy
 
 from qutebrowser.config import config, stylesheet
@@ -36,6 +36,7 @@ class Message(QLabel):
         super().__init__(text, parent)
         self.replace = replace
         self.setAttribute(Qt.WA_StyledBackground, True)
+        self.setWordWrap(True)
         qss = """
             padding-top: 2px;
             padding-bottom: 2px;
@@ -64,8 +65,6 @@ class Message(QLabel):
             """
         else:  # pragma: no cover
             raise ValueError("Invalid level {!r}".format(level))
-        # We don't bother with set_register_stylesheet here as it's short-lived
-        # anyways.
         stylesheet.set_register(self, qss, update=False)
 
 
@@ -77,7 +76,7 @@ class MessageView(QWidget):
 
     def __init__(self, parent=None):
         super().__init__(parent)
-        self._messages = []  # type: typing.MutableSequence[Message]
+        self._messages: MutableSequence[Message] = []
         self._vbox = QVBoxLayout(self)
         self._vbox.setContentsMargins(0, 0, 0, 0)
         self._vbox.setSpacing(0)
@@ -89,12 +88,6 @@ class MessageView(QWidget):
 
         self._last_text = None
 
-    def sizeHint(self):
-        """Get the proposed height for the view."""
-        height = sum(label.sizeHint().height() for label in self._messages)
-        # The width isn't really relevant as we're expanding anyways.
-        return QSize(-1, height)
-
     @config.change_filter('messages.timeout')
     def _set_clear_timer_interval(self):
         """Configure self._clear_timer according to the config."""
