@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,35 +15,46 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """Handlers for crashes and OS signals."""
 
 import os
+import os.path
 import sys
 import bdb
-import pdb  # flake8: disable=T002
+import pdb  # noqa: T002
 import signal
+import argparse
 import functools
+import threading
 import faulthandler
-import os.path
-import collections
+import dataclasses
+from typing import TYPE_CHECKING, Optional, MutableMapping, cast, List
 
 from PyQt5.QtCore import (pyqtSlot, qInstallMessageHandler, QObject,
                           QSocketNotifier, QTimer, QUrl)
-from PyQt5.QtWidgets import QApplication, QDialog
+from PyQt5.QtWidgets import QApplication
 
-from qutebrowser.commands import cmdutils
-from qutebrowser.misc import earlyinit, crashdialog
-from qutebrowser.utils import usertypes, standarddir, log, objreg, debug
+from qutebrowser.api import cmdutils
+from qutebrowser.misc import earlyinit, crashdialog, ipc, objects
+from qutebrowser.utils import usertypes, standarddir, log, objreg, debug, utils
+from qutebrowser.qt import sip
+if TYPE_CHECKING:
+    from qutebrowser.misc import quitter
 
 
-ExceptionInfo = collections.namedtuple('ExceptionInfo',
-                                       'pages, cmd_history, objects')
+@dataclasses.dataclass
+class ExceptionInfo:
+
+    """Information stored when there was an exception."""
+
+    pages: List[List[str]]
+    cmd_history: List[str]
+    objects: str
 
 
-# Used by mainwindow.py to skip confirm questions on crashes
-is_crashing = False
+crash_handler = cast('CrashHandler', None)
 
 
 class CrashHandler(QObject):
@@ -56,6 +67,9 @@ class CrashHandler(QObject):
         _args: The argparse namespace.
         _crash_dialog: The CrashDialog currently being shown.
         _crash_log_file: The file handle for the faulthandler crash log.
+        _crash_log_data: Crash data read from the previous crash log.
+        is_crashing: Used by mainwindow.py to skip confirm questions on
+                     crashes.
     """
 
     def __init__(self, *, app, quitter, args, parent=None):
@@ -64,28 +78,24 @@ class CrashHandler(QObject):
         self._quitter = quitter
         self._args = args
         self._crash_log_file = None
+        self._crash_log_data = None
         self._crash_dialog = None
+        self.is_crashing = False
 
     def activate(self):
         """Activate the exception hook."""
         sys.excepthook = self.exception_hook
 
-    def handle_segfault(self):
-        """Handle a segfault from a previous run."""
+    def init_faulthandler(self):
+        """Handle a segfault from a previous run and set up faulthandler."""
         logname = os.path.join(standarddir.data(), 'crash.log')
         try:
             # First check if an old logfile exists.
             if os.path.exists(logname):
                 with open(logname, 'r', encoding='ascii') as f:
-                    data = f.read()
+                    self._crash_log_data = f.read()
                 os.remove(logname)
                 self._init_crashlogfile()
-                if data:
-                    # Crashlog exists and has data in it, so something crashed
-                    # previously.
-                    self._crash_dialog = crashdialog.get_fatal_crash_dialog(
-                        self._args.debug, data)
-                    self._crash_dialog.show()
             else:
                 # There's no log file, so we can use this to display crashes to
                 # the user on the next start.
@@ -94,6 +104,17 @@ class CrashHandler(QObject):
             log.init.exception("Error while handling crash log file!")
             self._init_crashlogfile()
 
+    def display_faulthandler(self):
+        """If there was data in the crash log file, display a dialog."""
+        assert not self._args.no_err_windows
+        if self._crash_log_data:
+            # Crashlog exists and has data in it, so something crashed
+            # previously.
+            self._crash_dialog = crashdialog.FatalCrashDialog(
+                self._args.debug, self._crash_log_data)
+            self._crash_dialog.show()
+        self._crash_log_data = None
+
     def _recover_pages(self, forgiving=False):
         """Try to recover all open pages.
 
@@ -127,7 +148,6 @@ class CrashHandler(QObject):
 
     def _init_crashlogfile(self):
         """Start a new logfile and redirect faulthandler to it."""
-        assert not self._args.no_err_windows
         logname = os.path.join(standarddir.data(), 'crash.log')
         try:
             self._crash_log_file = open(logname, 'w', encoding='ascii')
@@ -137,14 +157,29 @@ class CrashHandler(QObject):
             earlyinit.init_faulthandler(self._crash_log_file)
 
     @cmdutils.register(instance='crash-handler')
-    def report(self):
-        """Report a bug in qutebrowser."""
+    def report(self, info=None, contact=None):
+        """Report a bug in qutebrowser.
+
+        Args:
+            info: Information about the bug report. If given, no report dialog
+                  shows up.
+            contact: Contact information for the report.
+        """
         pages = self._recover_pages()
         cmd_history = objreg.get('command-history')[-5:]
-        objects = debug.get_all_objects()
+        all_objects = debug.get_all_objects()
+
         self._crash_dialog = crashdialog.ReportDialog(pages, cmd_history,
-                                                      objects)
-        self._crash_dialog.show()
+                                                      all_objects)
+
+        if info is None:
+            self._crash_dialog.show()
+        else:
+            self._crash_dialog.report(info=info, contact=contact)
+
+    @pyqtSlot()
+    def shutdown(self):
+        self.destroy_crashlogfile()
 
     def destroy_crashlogfile(self):
         """Clean up the crash log file and delete it."""
@@ -156,7 +191,7 @@ class CrashHandler(QObject):
         if sys.__stderr__ is not None:
             faulthandler.enable(sys.__stderr__)
         else:
-            faulthandler.disable()
+            faulthandler.disable()  # type: ignore[unreachable]
         try:
             self._crash_log_file.close()
             os.remove(self._crash_log_file.name)
@@ -167,7 +202,7 @@ class CrashHandler(QObject):
         """Get info needed for the exception hook/dialog.
 
         Return:
-            An ExceptionInfo namedtuple.
+            An ExceptionInfo object.
         """
         try:
             pages = self._recover_pages(forgiving=True)
@@ -182,11 +217,47 @@ class CrashHandler(QObject):
             cmd_history = []
 
         try:
-            objects = debug.get_all_objects()
+            all_objects = debug.get_all_objects()
         except Exception:
             log.destroy.exception("Error while getting objects")
-            objects = ""
-        return ExceptionInfo(pages, cmd_history, objects)
+            all_objects = ""
+        return ExceptionInfo(pages, cmd_history, all_objects)
+
+    def _handle_early_exits(self, exc):
+        """Handle some special cases for the exception hook.
+
+        Return value:
+            True: Exception hook should be aborted.
+            False: Continue handling exception.
+        """
+        exctype, _excvalue, tb = exc
+
+        if not self._quitter.quit_status['crash']:
+            log.misc.error("ARGH, there was an exception while the crash "
+                           "dialog is already shown:", exc_info=exc)
+            return True
+
+        log.misc.error("Uncaught exception", exc_info=exc)
+
+        is_ignored_exception = (exctype is bdb.BdbQuit or
+                                not issubclass(exctype, Exception))
+
+        if 'pdb-postmortem' in objects.debug_flags:
+            if tb is None:
+                pdb.set_trace()  # noqa: T100
+            else:
+                pdb.post_mortem(tb)
+
+        if is_ignored_exception or 'pdb-postmortem' in objects.debug_flags:
+            # pdb exit, KeyboardInterrupt, ...
+            sys.exit(usertypes.Exit.exception)
+
+        if threading.current_thread() != threading.main_thread():
+            log.misc.error("Ignoring exception outside of main thread... "
+                           "Please report this as a bug.")
+            return True
+
+        return False
 
     def exception_hook(self, exctype, excvalue, tb):
         """Handle uncaught python exceptions.
@@ -195,39 +266,18 @@ class CrashHandler(QObject):
         gracefully.
         """
         exc = (exctype, excvalue, tb)
-        qapp = QApplication.instance()
 
-        if not self._quitter.quit_status['crash']:
-            log.misc.error("ARGH, there was an exception while the crash "
-                           "dialog is already shown:", exc_info=exc)
+        if self._handle_early_exits(exc):
             return
 
-        log.misc.error("Uncaught exception", exc_info=exc)
-
-        is_ignored_exception = (exctype is bdb.BdbQuit or
-                                not issubclass(exctype, Exception))
-
-        if 'pdb-postmortem' in self._args.debug_flags:
-            pdb.post_mortem(tb)
-
-        if is_ignored_exception or 'pdb-postmortem' in self._args.debug_flags:
-            # pdb exit, KeyboardInterrupt, ...
-            status = 0 if is_ignored_exception else 2
-            try:
-                self._quitter.shutdown(status)
-                return
-            except Exception:
-                log.init.exception("Error while shutting down")
-                qapp.quit()
-                return
-
         self._quitter.quit_status['crash'] = False
         info = self._get_exception_info()
 
-        try:
-            objreg.get('ipc-server').ignored = True
-        except Exception:
-            log.destroy.exception("Error while ignoring ipc")
+        if ipc.server is not None:
+            try:
+                ipc.server.ignored = True
+            except Exception:
+                log.destroy.exception("Error while ignoring ipc")
 
         try:
             self._app.lastWindowClosed.disconnect(
@@ -235,8 +285,7 @@ class CrashHandler(QObject):
         except TypeError:
             log.destroy.exception("Error while preventing shutdown")
 
-        global is_crashing
-        is_crashing = True
+        self.is_crashing = True
 
         self._app.closeAllWindows()
         if self._args.no_err_windows:
@@ -246,8 +295,8 @@ class CrashHandler(QObject):
             self._crash_dialog = crashdialog.ExceptionCrashDialog(
                 self._args.debug, info.pages, info.cmd_history, exc,
                 info.objects)
-            ret = self._crash_dialog.exec_()
-            if ret == QDialog.Accepted:  # restore
+            ret = self._crash_dialog.exec()
+            if ret == crashdialog.Result.restore:
                 self._quitter.restart(info.pages)
 
         # We might risk a segfault here, but that's better than continuing to
@@ -283,9 +332,9 @@ class SignalHandler(QObject):
         self._quitter = quitter
         self._notifier = None
         self._timer = usertypes.Timer(self, 'python_hacks')
-        self._orig_handlers = {}
+        self._orig_handlers: MutableMapping[int, 'signal._HANDLER'] = {}
         self._activated = False
-        self._orig_wakeup_fd = None
+        self._orig_wakeup_fd: Optional[int] = None
 
     def activate(self):
         """Set up signal handlers.
@@ -301,17 +350,20 @@ class SignalHandler(QObject):
         self._orig_handlers[signal.SIGTERM] = signal.signal(
             signal.SIGTERM, self.interrupt)
 
-        if os.name == 'posix' and hasattr(signal, 'set_wakeup_fd'):
+        if utils.is_posix and hasattr(signal, 'set_wakeup_fd'):
             # pylint: disable=import-error,no-member,useless-suppression
             import fcntl
             read_fd, write_fd = os.pipe()
             for fd in [read_fd, write_fd]:
                 flags = fcntl.fcntl(fd, fcntl.F_GETFL)
                 fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)
-            self._notifier = QSocketNotifier(read_fd, QSocketNotifier.Read,
+            self._notifier = QSocketNotifier(cast(sip.voidptr, read_fd),
+                                             QSocketNotifier.Read,
                                              self)
-            self._notifier.activated.connect(self.handle_signal_wakeup)
+            self._notifier.activated.connect(  # type: ignore[attr-defined]
+                self.handle_signal_wakeup)
             self._orig_wakeup_fd = signal.set_wakeup_fd(write_fd)
+            # pylint: enable=import-error,no-member,useless-suppression
         else:
             self._timer.start(1000)
             self._timer.timeout.connect(lambda: None)
@@ -322,10 +374,11 @@ class SignalHandler(QObject):
         if not self._activated:
             return
         if self._notifier is not None:
+            assert self._orig_wakeup_fd is not None
             self._notifier.setEnabled(False)
             rfd = self._notifier.socket()
             wfd = signal.set_wakeup_fd(self._orig_wakeup_fd)
-            os.close(rfd)
+            os.close(int(rfd))
             os.close(wfd)
         for sig, handler in self._orig_handlers.items():
             signal.signal(sig, handler)
@@ -340,11 +393,12 @@ class SignalHandler(QObject):
 
         Python will get control here, so the signal will get handled.
         """
+        assert self._notifier is not None
         log.destroy.debug("Handling signal wakeup!")
         self._notifier.setEnabled(False)
         read_fd = self._notifier.socket()
         try:
-            os.read(read_fd, 1)
+            os.read(int(read_fd), 1)
         except OSError:
             log.destroy.exception("Failed to read wakeup fd.")
         self._notifier.setEnabled(True)
@@ -390,3 +444,19 @@ class SignalHandler(QObject):
         """
         print("WHY ARE YOU DOING THIS TO ME? :(")
         sys.exit(128 + signum)
+
+
+def init(q_app: QApplication,
+         args: argparse.Namespace,
+         quitter: 'quitter.Quitter') -> None:
+    """Initialize crash/signal handlers."""
+    global crash_handler
+    crash_handler = CrashHandler(
+        app=q_app, quitter=quitter, args=args, parent=q_app)
+    objreg.register('crash-handler', crash_handler, command_only=True)
+    crash_handler.activate()
+    quitter.shutting_down.connect(crash_handler.shutdown)
+
+    signal_handler = SignalHandler(app=q_app, quitter=quitter, parent=q_app)
+    signal_handler.activate()
+    quitter.shutting_down.connect(signal_handler.deactivate)
