@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2014-2017 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2014-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,43 +15,38 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
-"""Custom useful data types.
-
-Module attributes:
-    _UNSET: Used as default argument in the constructor so default can be None.
-"""
+"""Custom useful data types."""
 
 import operator
-import collections.abc
-import enum as pyenum
+import enum
+import dataclasses
+from typing import Any, Optional, Sequence, TypeVar, Union
 
 from PyQt5.QtCore import pyqtSignal, pyqtSlot, QObject, QTimer
+from PyQt5.QtCore import QUrl
 
 from qutebrowser.utils import log, qtutils, utils
 
 
-_UNSET = object()
+_T = TypeVar('_T', bound=utils.Comparable)
 
 
-def enum(name, items, start=1, is_int=False):
-    """Factory for simple enumerations.
+class Unset:
 
-    Args:
-        name: Name of the enum
-        items: Iterable of items to be sequentially enumerated.
-        start: The number to use for the first value.
-               We use 1 as default so enum members are always True.
-        is_init: True if the enum should be a Python IntEnum
-    """
-    enums = [(v, i) for (i, v) in enumerate(items, start)]
-    base = pyenum.IntEnum if is_int else pyenum.Enum
-    base = pyenum.unique(base)
-    return base(name, enums)
+    """Class for an unset object."""
+
+    __slots__ = ()
+
+    def __repr__(self) -> str:
+        return '<UNSET>'
 
 
-class NeighborList(collections.abc.Sequence):
+UNSET = Unset()
+
+
+class NeighborList(Sequence[_T]):
 
     """A list of items which saves its current position.
 
@@ -65,9 +60,16 @@ class NeighborList(collections.abc.Sequence):
         _mode: The current mode.
     """
 
-    Modes = enum('Modes', ['edge', 'exception'])
+    class Modes(enum.Enum):
 
-    def __init__(self, items=None, default=_UNSET, mode=Modes.exception):
+        """Behavior for the 'mode' argument."""
+
+        edge = enum.auto()
+        exception = enum.auto()
+
+    def __init__(self, items: Sequence[_T] = None,
+                 default: Union[_T, Unset] = UNSET,
+                 mode: Modes = Modes.exception) -> None:
         """Constructor.
 
         Args:
@@ -80,28 +82,31 @@ class NeighborList(collections.abc.Sequence):
         if not isinstance(mode, self.Modes):
             raise TypeError("Mode {} is not a Modes member!".format(mode))
         if items is None:
-            self._items = []
+            self._items: Sequence[_T] = []
         else:
             self._items = list(items)
         self._default = default
-        if default is not _UNSET:
-            self._idx = self._items.index(default)
+
+        if not isinstance(default, Unset):
+            idx = self._items.index(default)
+            self._idx: Optional[int] = idx
         else:
             self._idx = None
-        self._mode = mode
-        self.fuzzyval = None
 
-    def __getitem__(self, key):
+        self._mode = mode
+        self.fuzzyval: Optional[int] = None
+
+    def __getitem__(self, key: int) -> _T:  # type: ignore[override]
         return self._items[key]
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self._items)
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return utils.get_repr(self, items=self._items, mode=self._mode,
                               idx=self._idx, fuzzyval=self.fuzzyval)
 
-    def _snap_in(self, offset):
+    def _snap_in(self, offset: int) -> bool:
         """Set the current item to the closest item to self.fuzzyval.
 
         Args:
@@ -112,20 +117,24 @@ class NeighborList(collections.abc.Sequence):
             True if the value snapped in (changed),
             False when the value already was in the list.
         """
+        assert isinstance(self.fuzzyval, (int, float)), self.fuzzyval
+
         op = operator.le if offset < 0 else operator.ge
         items = [(idx, e) for (idx, e) in enumerate(self._items)
                  if op(e, self.fuzzyval)]
         if items:
-            item = min(items, key=lambda tpl: abs(self.fuzzyval - tpl[1]))
+            item = min(
+                items,
+                key=lambda tpl:
+                abs(self.fuzzyval - tpl[1]))  # type: ignore[operator]
         else:
-            sorted_items = sorted([(idx, e) for (idx, e) in
-                                   enumerate(self.items)], key=lambda e: e[1])
+            sorted_items = sorted(enumerate(self.items), key=lambda e: e[1])
             idx = 0 if offset < 0 else -1
             item = sorted_items[idx]
         self._idx = item[0]
         return self.fuzzyval not in self._items
 
-    def _get_new_item(self, offset):
+    def _get_new_item(self, offset: int) -> _T:
         """Logic for getitem to get the item at offset.
 
         Args:
@@ -134,6 +143,7 @@ class NeighborList(collections.abc.Sequence):
         Return:
             The new item.
         """
+        assert self._idx is not None
         try:
             if self._idx + offset >= 0:
                 new = self._items[self._idx + offset]
@@ -153,11 +163,11 @@ class NeighborList(collections.abc.Sequence):
         return new
 
     @property
-    def items(self):
+    def items(self) -> Sequence[_T]:
         """Getter for items, which should not be set."""
         return self._items
 
-    def getitem(self, offset):
+    def getitem(self, offset: int) -> _T:
         """Get the item with a relative position.
 
         Args:
@@ -182,86 +192,168 @@ class NeighborList(collections.abc.Sequence):
             self.fuzzyval = None
         return self._get_new_item(offset)
 
-    def curitem(self):
+    def curitem(self) -> _T:
         """Get the current item in the list."""
         if self._idx is not None:
             return self._items[self._idx]
         else:
             raise IndexError("No current item!")
 
-    def nextitem(self):
+    def nextitem(self) -> _T:
         """Get the next item in the list."""
         return self.getitem(1)
 
-    def previtem(self):
+    def previtem(self) -> _T:
         """Get the previous item in the list."""
         return self.getitem(-1)
 
-    def firstitem(self):
+    def firstitem(self) -> _T:
         """Get the first item in the list."""
         if not self._items:
             raise IndexError("No items found!")
         self._idx = 0
         return self.curitem()
 
-    def lastitem(self):
+    def lastitem(self) -> _T:
         """Get the last item in the list."""
         if not self._items:
             raise IndexError("No items found!")
         self._idx = len(self._items) - 1
         return self.curitem()
 
-    def reset(self):
+    def reset(self) -> _T:
         """Reset the position to the default."""
-        if self._default is _UNSET:
+        if self._default is UNSET:
             raise ValueError("No default set!")
-        else:
-            self._idx = self._items.index(self._default)
-            return self.curitem()
+        self._idx = self._items.index(self._default)
+        return self.curitem()
 
 
-# The mode of a Question.
-PromptMode = enum('PromptMode', ['yesno', 'text', 'user_pwd', 'alert',
-                                 'download'])
+class PromptMode(enum.Enum):
+
+    """The mode of a Question."""
+
+    yesno = enum.auto()
+    text = enum.auto()
+    user_pwd = enum.auto()
+    alert = enum.auto()
+    download = enum.auto()
 
 
-# Where to open a clicked link.
-ClickTarget = enum('ClickTarget', ['normal', 'tab', 'tab_bg', 'window',
-                                   'hover'])
+class ClickTarget(enum.Enum):
+
+    """How to open a clicked link."""
+
+    normal = enum.auto()  #: Open the link in the current tab
+    tab = enum.auto()  #: Open the link in a new foreground tab
+    tab_bg = enum.auto()  #: Open the link in a new background tab
+    window = enum.auto()  #: Open the link in a new window
+    hover = enum.auto()  #: Only hover over the link
 
 
-# Key input modes
-KeyMode = enum('KeyMode', ['normal', 'hint', 'command', 'yesno', 'prompt',
-                           'insert', 'passthrough', 'caret', 'set_mark',
-                           'jump_mark', 'record_macro', 'run_macro'])
+class KeyMode(enum.Enum):
+
+    """Key input modes."""
+
+    normal = enum.auto()  #: Normal mode (no mode was entered)
+    hint = enum.auto()  #: Hint mode (showing labels for links)
+    command = enum.auto()  #: Command mode (after pressing the colon key)
+    yesno = enum.auto()  #: Yes/No prompts
+    prompt = enum.auto()  #: Text prompts
+    insert = enum.auto()  #: Insert mode (passing through most keys)
+    passthrough = enum.auto()  #: Passthrough mode (passing through all keys)
+    caret = enum.auto()  #: Caret mode (moving cursor with keys)
+    set_mark = enum.auto()
+    jump_mark = enum.auto()
+    record_macro = enum.auto()
+    run_macro = enum.auto()
+    # 'register' is a bit of an oddball here: It's not really a "real" mode,
+    # but it's used in the config for common bindings for
+    # set_mark/jump_mark/record_macro/run_macro.
+    register = enum.auto()
 
 
-# Available command completions
-Completion = enum('Completion', ['command', 'section', 'option', 'value',
-                                 'helptopic', 'quickmark_by_name',
-                                 'bookmark_by_url', 'url', 'tab', 'sessions',
-                                 'bind'])
+class Exit(enum.IntEnum):
+
+    """Exit statuses for errors. Needs to be an int for sys.exit."""
+
+    ok = 0
+    reserved = 1
+    exception = 2
+    err_ipc = 3
+    err_init = 4
 
 
-# Exit statuses for errors. Needs to be an int for sys.exit.
-Exit = enum('Exit', ['ok', 'reserved', 'exception', 'err_ipc', 'err_init',
-                     'err_config', 'err_key_config'], is_int=True, start=0)
+class LoadStatus(enum.Enum):
+
+    """Load status of a tab."""
+
+    none = enum.auto()
+    success = enum.auto()
+    success_https = enum.auto()
+    error = enum.auto()
+    warn = enum.auto()
+    loading = enum.auto()
 
 
-# Load status of a tab
-LoadStatus = enum('LoadStatus', ['none', 'success', 'success_https', 'error',
-                                 'warn', 'loading'])
+class Backend(enum.Enum):
+
+    """The backend being used (usertypes.backend)."""
+
+    QtWebKit = enum.auto()
+    QtWebEngine = enum.auto()
 
 
-# Backend of a tab
-Backend = enum('Backend', ['QtWebKit', 'QtWebEngine'])
+class JsWorld(enum.Enum):
+
+    """World/context to run JavaScript code in."""
+
+    main = enum.auto()  #: Same world as the web page's JavaScript.
+    application = enum.auto()  #: Application world, used by qutebrowser internally.
+    user = enum.auto()  #: User world, currently not used.
+    jseval = enum.auto()  #: World used for the jseval-command.
 
 
-# JS world for QtWebEngine
-JsWorld = enum('JsWorld', ['main', 'application', 'user', 'jseval'])
+class JsLogLevel(enum.Enum):
+
+    """Log level of a JS message.
+
+    This needs to match up with the keys allowed for the
+    content.javascript.log setting.
+    """
+
+    unknown = enum.auto()
+    info = enum.auto()
+    warning = enum.auto()
+    error = enum.auto()
 
 
-MessageLevel = enum('MessageLevel', ['error', 'warning', 'info'])
+class MessageLevel(enum.Enum):
+
+    """The level of a message being shown."""
+
+    error = enum.auto()
+    warning = enum.auto()
+    info = enum.auto()
+
+
+class IgnoreCase(enum.Enum):
+
+    """Possible values for the 'search.ignore_case' setting."""
+
+    smart = enum.auto()
+    never = enum.auto()
+    always = enum.auto()
+
+
+class CommandValue(enum.Enum):
+
+    """Special values which are injected when running a command handler."""
+
+    count = enum.auto()
+    win_id = enum.auto()
+    cur_tab = enum.auto()
+    count_tab = enum.auto()
 
 
 class Question(QObject):
@@ -283,6 +375,8 @@ class Question(QObject):
                  For user_pwd, a default username as string.
         title: The question title to show.
         text: The prompt text to display to the user.
+        url: Any URL referenced in prompts.
+        option: Boolean option to be set when answering always/never.
         answer: The value the user entered (as password for user_pwd).
         is_aborted: Whether the question was aborted.
         interrupted: Whether the question was interrupted by another one.
@@ -307,34 +401,25 @@ class Question(QObject):
     answered_no = pyqtSignal()
     completed = pyqtSignal()
 
-    def __init__(self, parent=None):
+    def __init__(self, parent: QObject = None) -> None:
         super().__init__(parent)
-        self._mode = None
-        self.default = None
-        self.title = None
-        self.text = None
-        self.answer = None
+        self.mode: Optional[PromptMode] = None
+        self.default: Union[bool, str, None] = None
+        self.title: Optional[str] = None
+        self.text: Optional[str] = None
+        self.url: Optional[str] = None
+        self.option: Optional[bool] = None
+        self.answer: Union[str, bool, None] = None
         self.is_aborted = False
         self.interrupted = False
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return utils.get_repr(self, title=self.title, text=self.text,
-                              mode=self._mode, default=self.default)
-
-    @property
-    def mode(self):
-        """Getter for mode so we can define a setter."""
-        return self._mode
-
-    @mode.setter
-    def mode(self, val):
-        """Setter for mode to do basic type checking."""
-        if not isinstance(val, PromptMode):
-            raise TypeError("Mode {} is no PromptMode member!".format(val))
-        self._mode = val
+                              mode=self.mode, default=self.default,
+                              option=self.option)
 
     @pyqtSlot()
-    def done(self):
+    def done(self) -> None:
         """Must be called when the question was answered completely."""
         self.answered.emit(self.answer)
         if self.mode == PromptMode.yesno:
@@ -345,26 +430,20 @@ class Question(QObject):
         self.completed.emit()
 
     @pyqtSlot()
-    def cancel(self):
+    def cancel(self) -> None:
         """Cancel the question (resulting from user-input)."""
         self.cancelled.emit()
         self.completed.emit()
 
     @pyqtSlot()
-    def abort(self):
+    def abort(self) -> None:
         """Abort the question."""
         if self.is_aborted:
             log.misc.debug("Question was already aborted")
             return
         self.is_aborted = True
-        try:
-            self.aborted.emit()
-            self.completed.emit()
-        except TypeError:
-            # WORKAROUND
-            # We seem to get "pyqtSignal must be bound to a QObject, not
-            # 'Question' here, which makes no sense at all..."
-            log.misc.exception("Error while aborting question")
+        self.aborted.emit()
+        self.completed.emit()
 
 
 class Timer(QTimer):
@@ -375,7 +454,7 @@ class Timer(QTimer):
         _name: The name of the timer.
     """
 
-    def __init__(self, parent=None, name=None):
+    def __init__(self, parent: QObject = None, name: str = None) -> None:
         super().__init__(parent)
         if name is None:
             self._name = "unnamed"
@@ -383,15 +462,15 @@ class Timer(QTimer):
             self.setObjectName(name)
             self._name = name
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return utils.get_repr(self, name=self._name)
 
-    def setInterval(self, msec):
+    def setInterval(self, msec: int) -> None:
         """Extend setInterval to check for overflows."""
         qtutils.check_overflow(msec, 'int')
         super().setInterval(msec)
 
-    def start(self, msec=None):
+    def start(self, msec: int = None) -> None:
         """Extend start to check for overflows."""
         if msec is not None:
             qtutils.check_overflow(msec, 'int')
@@ -404,14 +483,50 @@ class AbstractCertificateErrorWrapper:
 
     """A wrapper over an SSL/certificate error."""
 
-    def __init__(self, error):
+    def __init__(self, error: Any) -> None:
         self._error = error
 
-    def __str__(self):
+    def __str__(self) -> str:
         raise NotImplementedError
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         raise NotImplementedError
 
-    def is_overridable(self):
+    def is_overridable(self) -> bool:
         raise NotImplementedError
+
+
+@dataclasses.dataclass
+class NavigationRequest:
+
+    """A request to navigate to the given URL."""
+
+    class Type(enum.Enum):
+
+        """The type of a request.
+
+        Based on QWebEngineUrlRequestInfo::NavigationType and QWebPage::NavigationType.
+        """
+
+        #: Navigation initiated by clicking a link.
+        link_clicked = 1
+        #: Navigation explicitly initiated by typing a URL (QtWebEngine only).
+        typed = 2
+        #: Navigation submits a form.
+        form_submitted = 3
+        #: An HTML form was submitted a second time (QtWebKit only).
+        form_resubmitted = 4
+        #: Navigation initiated by a history action.
+        back_forward = 5
+        #: Navigation initiated by refreshing the page.
+        reloaded = 6
+        #: Navigation triggered automatically by page content or remote server
+        #: (QtWebEngine >= 5.14 only)
+        redirect = 7
+        #: None of the above.
+        other = 8
+
+    url: QUrl
+    navigation_type: Type
+    is_main_frame: bool
+    accepted: bool = True
